/* $Workfile: Arni6x6Demo.c $       $Revision: 28 $      $Modtime: 10/21/05 6:06p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    Arni6x6Demo.c

    A basic example console application demonstrating the use of OTX Driver
    with the Arni-6x6-PCI or the Arni-6x6-PCI board.

    The sample application configures 6 FXO POTS Line Interfaces.
    The application provides a simple console interface to manage the POTS 
    Line Interface (LI) devices.
   
    The user can use the following simple keystrokes as commands to go off-hook
    and on-hook, to dial numbers, etc.

    In this basic demo the FXO POTS LIs are supposed to be connected to a telephone 
    line or a telephone line simulator.
    
    The main command keys are listed in the following:
     - Press 'h' or '?' for help. 'q' to quit.
     

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxBrd.h"
#include "OtxCst.h"
#include "OtxDsp.h"
#include "OtxBurst.h"
#include "OtxBurstD.h"



/*******************************************************************************
  Macros, Types
*******************************************************************************/

#define DEMO_DSP_MAX 1
#define DEMO_LI_MAX 1

#define DEMO_TASK_REF           1L                      // Task Reference for a non-blocking call
#define DEMO_DEFAULT_LOG_FILE   "Arni6x6Demo.log"       // Name of the LOG file
#define DEMO_DEFAULT_BOARD_NO   0
#define DEMO_DEFAULT_FXO_NO     2

#define DEMO_BURST_BUF_SIZE     0x8000    // 32k

#define DEMO_U_LAW_SILENCE   0xFF    // Sending this byte produces silence in u-law mode
#define DEMO_A_LAW_SILENCE   0xD5    // Sending this byte produces silence in a-law mode

#define DEMO_HWY_DSP0   0
#define DEMO_HWY_DSP1   1
#define DEMO_HWY_DSP2   2
#define DEMO_HWY_DSP3   3
#define DEMO_HWY_DSP4   4
#define DEMO_HWY_DSP5   5
#define DEMO_HWY_BURST  8


typedef struct DemoEnvT {
  char           *pszInputStr;
  OTX_BOOL        bExitOnError;
  char           *logFileName;
  char           *outFileName;
  int             boardNo;
  int             fxoNo;
  OtxPhysicalDeviceTypeE  nBoardType;
  OtxLogicalDeviceTypeE  eDetectorType;
  OtxDataFormatE  law;
  OtxCstPotsCfgSetE  nSetId;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static char DemoDtmfParamToChar(int nParam);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static OTX_INT32 DemoProcessUserCommand(char commandChar);
static int DemoExit(int nExitCode);
static void DemoPrintPresentationMessage(OtxCstCidPresentationMessageS *pCidMsg);

/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue        = OTX_INVALID_HANDLE_VALUE;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
FILE       *g_fStreamFile = NULL;
OTX_BOOL    g_bWatchDogThreadActive = OTX_FALSE;
OTX_BOOL    g_bKickTheWatchDog = OTX_FALSE;
OTX_BOOL    g_bLineSensingThreadActive = OTX_FALSE;
OTX_BOOL    g_bLineSensePending = OTX_FALSE;
OTX_BOOL    g_bLineSenseVerbose = OTX_FALSE;
OTX_INT32   g_nLineSenseLastValue = -1;
OTX_BOOL    g_bLibConnected = OTX_FALSE;

static OsaMutexS g_demoLogfCritSec;

struct {  // Handles to Physical Devices
  OTX_HANDLE hBoardController0;
  OTX_HANDLE hDsp[6];
  OTX_HANDLE hBurst;
  OTX_HANDLE hTss;
} g_hPhysDevs;


// direct Burst DMA send
struct sBurstSender {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  OTX_UINT8   sqTxBuf[DEMO_BURST_BUF_SIZE];
  FILE        *pTxFile;
  OTX_INT32   nTsToSend;  
  OTX_UINT32  cbTxBuf;
  OTX_BOOL    bFileLoop;
};

// direct Burst DMA receive
struct sBurstReceiver {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  FILE        *pRxFile;
  OTX_INT32   nTsToRecord;  
};

// Pots Li device
struct sPotsLi {
  OTX_BOOL    bOffHook;
};


struct {  // Handles to Logical Devices
  OTX_HANDLE hPotsLi[DEMO_LI_MAX];
  OTX_HANDLE hVaDetector[DEMO_LI_MAX];
  OTX_HANDLE hCidDetector[DEMO_LI_MAX];
  OTX_HANDLE hCptDetector[DEMO_LI_MAX];
  OTX_HANDLE hDtmfDetector[DEMO_LI_MAX];
  OTX_HANDLE hDtmfDialer[DEMO_LI_MAX];
  OTX_HANDLE hVoiceMixer[DEMO_LI_MAX];
  OTX_HANDLE hUserApp[DEMO_LI_MAX];     // only used for debugging of the DSP program
} g_hLogicDevs;

struct {
  struct sBurstSender mBurstSender;        
  struct sBurstReceiver mBurstReceiver;    
} g_sBurstDevs;

struct {
  struct sPotsLi mPotsLi;        
} g_sPotsDevs[DEMO_LI_MAX];


struct {  // Handles to Pipes
  OTX_HANDLE hPotsRx[DEMO_LI_MAX];
  OTX_HANDLE hPotsTx[DEMO_LI_MAX];
  OTX_HANDLE hMixer0[DEMO_LI_MAX];
  OTX_HANDLE hMixer1[DEMO_LI_MAX];
} g_hPipes;

char *g_sqNumber0 = "102";

/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************
  @func
    Create the event queue to be used by the driver to inform the application
    about events, such as line status changes, or detected DMTF digits.
*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

}



void DemoOpenPhysicalDevices(void)
/*******************************************************************************
  @func
    Opens the Physical Devices on the Arni-16-PCI board

  @comm
    Calls function <f OtxDrvOpenPhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Get Handle to the Physical ARNI Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_ARNI_6X6_PCI_PLUS:
    DemoLogf("Opening Arni-6x6-PCI-Plus board\n");
    break;
  case OTX_DEVICE_ARNI_6X6_PCI:
    DemoLogf("Opening Arni-6x6-PCI board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    DemoExitQuestion();
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hBoardController0); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  
  // Check if the watchdog timer has triggered before we continue
  {
    OTX_UINT32 readVal;
    result = OtxDrvReadReg32(g_hPhysDevs.hBoardController0, 0x04, &readVal); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    if (readVal == 0) {
      DemoLogf("The watchdog timer has triggered. A reboot of the PC is necessary.\n");
      DemoExitQuestion();
    }
  }

  // Disable the Watchdog Timer, and enable it when we have completed the configuration
  result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  OTX_FALSE); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Get a handle to the Time-Space Switch (TSS) on the board (just for debugging purposes)
  DemoLogf("Open OTX_DEVICE_TSS\n");
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hPhysDevs.hTss); DemoCheckRc(g_hPhysDevs.hBoardController0, result);


  for (i = 0; i < DEMO_DSP_MAX; i++) {
    DemoLogf("Open OTX_DEVICE_DSP #%d\n", g_demoEnv.fxoNo);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_DSP, g_demoEnv.fxoNo, g_demoEnv.fxoNo, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

  for (i = 0; i < DEMO_DSP_MAX; i++) {
    if (g_demoEnv.outFileName) {
      DemoLogf("Loading program file %s to DSP#%d \n", g_demoEnv.outFileName, g_demoEnv.fxoNo);
      result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], g_demoEnv.outFileName, "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    } else {
      DemoLogf("Loading OTX_DSP_SPM_OTXCST2 to DSP#%d \n", g_demoEnv.fxoNo);
      result = OtxDspRunSpm(g_hPhysDevs.hDsp[i], OTX_DSP_SPM_OTXCST2); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }
  }

  // Get a handle to the physical Burst device.
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_BURST, 0, 0, g_hEventQueue, &g_hPhysDevs.hBurst); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // set device buffer event trigger levels
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,  DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,    DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);

}



void DemoEnablePhysicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Physical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  DemoLogf("In DemoEnablePhysicalDevices\n");
  result = OtxDrvEnable(g_hPhysDevs.hBoardController0,     OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  for (i = 0; i < DEMO_DSP_MAX; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }
    
}


/*******************************************************************************
  @func
    Create the needed Logical Devices

  @comm
    The logical devices are different from the phyiscal devices in that the
    the number of logical devices is not fixed. More of them can be requested
    by the user and allocated by the system if needed.
*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  nDsp;
  OTX_INT32  nLi;

  for (nLi = 0; nLi < DEMO_LI_MAX; nLi++) {
    
    nDsp = nLi;
    
    DemoLogf("Creating Logical Devices on Dsp%d\n", nDsp);
    
    if (g_demoEnv.outFileName) {
      // For debug prints
      result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nLi], OTX_LDEVICE_USER_APPLICATION, g_hEventQueue, nLi, &g_hLogicDevs.hUserApp[nLi]); 
      //DemoCheckRc(g_hPhysDevs.hDsp[nLi], result);
      DemoLogf("Created User App device [%d], Handle = 0x%X (for debug printfs)\n", nLi, g_hLogicDevs.hUserApp[nLi]);
    }
    
    // Create Pots Li Logical Devices
    //DemoLogf("Create Logical Device OTX_LDEVICE_CST_LI_POTS #%d (nDsp=%d)\n", nLi, nDsp);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_LI_POTS, g_hEventQueue, nLi, &g_hLogicDevs.hPotsLi[nLi]); DemoCheckRc(g_hLogicDevs.hPotsLi[nLi], result);

    // Create Voice Activity Detector Logical Devices
    //DemoLogf("Create Logical Device OTX_LDEVICE_CST_VOICE_ACTIVITY_DETECTOR #%d (nDsp=%d)\n", nLi, nDsp);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_VOICE_ACTIVITY_DETECTOR, g_hEventQueue, nLi, &g_hLogicDevs.hVaDetector[nLi]); DemoCheckRc(g_hLogicDevs.hVaDetector[nLi], result);

    // Creatr CID Detector devices
    //DemoLogf("Creating CID Detector device [%d]\n", i);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_CALLER_ID_DETECTOR, g_hEventQueue, nLi, &g_hLogicDevs.hCidDetector[nLi]); DemoCheckRc(g_hPhysDevs.hDsp[nLi], result);

    // Create Cpt Detector Logical Devices
    //DemoLogf("Create Logical Device OTX_LDEVICE_CST_CPT_DETECTOR #%d (nDsp=%d)\n", nLi, nDsp);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_CPT_DETECTOR, g_hEventQueue, nLi, &g_hLogicDevs.hCptDetector[nLi]); DemoCheckRc(g_hLogicDevs.hCptDetector[nLi], result);

    // Create Dtmf Detector Logical Devices
    //DemoLogf("Create Logical Device OTX_LDEVICE_CST_DTMF_DETECTOR #%d (nDsp=%d)\n", nLi, nDsp);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_DTMF_DETECTOR, g_hEventQueue, nLi, &g_hLogicDevs.hDtmfDetector[nLi]); DemoCheckRc(g_hLogicDevs.hDtmfDetector[nLi], result);

    // Create Voice Mixer Logical Devices
    //DemoLogf("Create Logical Device OTX_LDEVICE_CST_VOICE_MIXER #%d (nDsp=%d)\n", nLi, nDsp);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_VOICE_MIXER, g_hEventQueue, nLi, &g_hLogicDevs.hVoiceMixer[nLi]); DemoCheckRc(g_hPhysDevs.hDsp[nLi], result);

    
    // Create DTMF Dialers Logical Devices
    //DemoLogf("Create Logical Device OTX_LDEVICE_CST_DTMF_DIALER #%d (nDsp=%d)\n", nLi, nDsp);
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_DTMF_DIALER, g_hEventQueue, nLi, &g_hLogicDevs.hDtmfDialer[nLi]); DemoCheckRc(g_hLogicDevs.hDtmfDialer[nLi], result);
    
  } 
  
}


void DemoConnectDevices(
/*******************************************************************************
  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.
*******************************************************************************/
  void
)
{
  OTX_INT32 nDsp;
  OTX_RESULT result;
  OTX_UINT32 iHighwayNo;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 nTs;


  // Create the Pipes (one for each direction) between the Burst Device and the DSPs
  for (nDsp = 0; nDsp < DEMO_LI_MAX; nDsp++) {

    // Burst highway number [0]
    iHighwayNo = 0;

    // Create pipes (one for each direction)
    DemoLogf("Creating and connection pipes\n");

    // -------------------------
    // Connect pipe: Dsp -> Dsp (DTMF dialer)
    // -------------------------
    result = OtxDrvCreatePipe(64, &g_hPipes.hPotsTx[nDsp]);
    DemoCheckRc(g_hPipes.hPotsTx[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 1;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hPotsTx[nDsp], OTX_PIPE_INPUT, g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hPotsTx[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 2;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hPotsTx[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hPotsTx[nDsp], result);

    // -------------------------
    // Connect pipe: (Daa ->) Dsp --> Dsp (DTMF detector)
    // -------------------------
    result = OtxDrvCreatePipe(64, &g_hPipes.hPotsRx[nDsp]);
    DemoCheckRc(g_hPipes.hPotsRx[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 0;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hPotsRx[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hPotsRx[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 0;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hPotsRx[nDsp], OTX_PIPE_OUTPUT,  g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hPotsRx[nDsp], result);

    // -------------------------
    // Connect pipe: Dsp --> Dsp (Voice Mixer 0 input)
    // -------------------------
    result = OtxDrvCreatePipe(64, &g_hPipes.hMixer0[nDsp]);
    DemoCheckRc(g_hPipes.hMixer0[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 20;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hMixer0[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hMixer0[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 20;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hMixer0[nDsp], OTX_PIPE_OUTPUT,  g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hMixer0[nDsp], result);

    // -------------------------
    // Connect pipe: Burst -> Dsp (Voice Mixer 1 input)
    // -------------------------
    result = OtxDrvCreatePipe(64, &g_hPipes.hMixer1[nDsp]);
    DemoCheckRc(g_hPipes.hMixer1[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 0;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hMixer1[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hBurst, iHighwayNo, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hMixer1[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 21;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hMixer1[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hMixer1[nDsp], result);
  
  }

  for (nDsp = 0; nDsp < DEMO_LI_MAX; nDsp++) {

    // CONNECT LOGICAL DEVICES WITHIN THE DSP
    // --------------------------------------


    // Create the following Connections in receive direction:
    //   Pots Li  --> Pipe --->  DtmfDetector 
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hPotsLi[nDsp], -1, g_hPipes.hPotsRx[nDsp], -1); DemoCheckRc(g_hPipes.hPotsRx[nDsp], result);
    switch (g_demoEnv.eDetectorType) {
    case OTX_LDEVICE_CST_CPT_DETECTOR:
      result = OtxDrvConnectLogicalDevice(g_hPipes.hPotsRx[nDsp], -1, g_hLogicDevs.hCptDetector[nDsp], -1); DemoCheckRc(g_hPipes.hPotsRx[nDsp], result);
      break;
    case OTX_LDEVICE_CST_DTMF_DETECTOR:
      result = OtxDrvConnectLogicalDevice(g_hPipes.hPotsRx[nDsp], -1, g_hLogicDevs.hDtmfDetector[nDsp], -1); DemoCheckRc(g_hPipes.hPotsRx[nDsp], result);
      break;
    case OTX_LDEVICE_CST_VOICE_ACTIVITY_DETECTOR:
      result = OtxDrvConnectLogicalDevice(g_hPipes.hPotsRx[nDsp], -1, g_hLogicDevs.hVaDetector[nDsp], -1); DemoCheckRc(g_hPipes.hPotsRx[nDsp], result);
      break;
    case OTX_LDEVICE_CST_CALLER_ID_DETECTOR:
      result = OtxDrvConnectLogicalDevice(g_hPipes.hPotsRx[nDsp], -1, g_hLogicDevs.hCidDetector[nDsp], -1); DemoCheckRc(g_hPipes.hPotsRx[nDsp], result);
      break;
    default:
      DemoLogf("Error: Unknown detector type configured\n");
      break;
    }


    // Make the following connections in transmit direction:
    //   DtmfDialer  --> Pipe --> Pots Li
    result = OtxDrvConnectLogicalDevice(g_hPipes.hPotsTx[nDsp], -1, g_hLogicDevs.hPotsLi[nDsp], -1); DemoCheckRc(g_hLogicDevs.hDtmfDialer[nDsp], result);

    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hDtmfDialer[nDsp], -1, g_hPipes.hMixer0[nDsp], -1); DemoCheckRc(g_hLogicDevs.hDtmfDialer[nDsp], result);

    // Logically connect the one output from the Voice Mixer (to the DAA tx pipe)
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hVoiceMixer[nDsp], -1, g_hPipes.hPotsTx[nDsp], -1); DemoCheckRc(g_hLogicDevs.hDtmfDialer[nDsp], result);

    // Logically connect the two inputs to the Voice Mixer
    result = OtxDrvConnectLogicalDevice(g_hPipes.hMixer0[nDsp], -1, g_hLogicDevs.hVoiceMixer[nDsp], 0); DemoCheckRc(g_hLogicDevs.hVoiceMixer[nDsp], result);
    result = OtxDrvConnectLogicalDevice(g_hPipes.hMixer1[nDsp], -1, g_hLogicDevs.hVoiceMixer[nDsp], 1); DemoCheckRc(g_hLogicDevs.hVoiceMixer[nDsp], result);

  }

#if 1 // For additional debugging
  {
    OTX_HANDLE hPipeRecDaa;
    OTX_INT32 nHwy;
    
    // Also connect the DAA samples to the burst (TS0) via a TSS connection
    result = OtxDrvCreatePipe(64, &hPipeRecDaa);
    DemoCheckRc(hPipeRecDaa, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 0;
    nHwy = 2;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeRecDaa, OTX_PIPE_INPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeRecDaa, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 0;
    nHwy = 8;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeRecDaa, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeRecDaa, result);
  }
#endif

#if 1 // For additional debugging
  {
    OTX_HANDLE hPipeTemp;
    OTX_INT32 nHwy;
    
    // Also connect the output from the Voice Mixer samples to the burst (TS1) via a TSS connection
    result = OtxDrvCreatePipe(64, &hPipeTemp);
    DemoCheckRc(hPipeTemp, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 1;
    nHwy = 2;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeTemp, OTX_PIPE_INPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeTemp, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 1;
    nHwy = 8;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeTemp, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeTemp, result);
  }
#endif

#if 1 // For additional debugging
  {
    OTX_HANDLE hPipeTemp;
    OTX_INT32 nHwy;
    
    // Also connect the output from the DTMF Dialer samples to the burst (TS2) via a TSS connection
    result = OtxDrvCreatePipe(64, &hPipeTemp);
    DemoCheckRc(hPipeTemp, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 20;
    nHwy = 2;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeTemp, OTX_PIPE_INPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeTemp, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 2;
    nHwy = 8;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeTemp, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeTemp, result);
  }
#endif



#if 1 // For additional debugging
  {
    OTX_HANDLE hPipeTemp;
    OTX_INT32 nHwy;
    
    // Also connect the output from the Burst Sender samples to the burst (TS3) via a TSS connection
    result = OtxDrvCreatePipe(64, &hPipeTemp);
    DemoCheckRc(hPipeTemp, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 0;
    nHwy = 8;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeTemp, OTX_PIPE_INPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeTemp, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 3;
    nHwy = 8;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeTemp, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeTemp, result);
  }
#endif



#if 0 // Connect all TSs between the DSP and burst for debugging
  /*
  {
    OTX_HANDLE hPipe;
    OTX_RESULT result;
    OTX_UINT8 sqTimeSlotMask[32];
    OTX_INT32 nTs;
    
    nDsp = 0;
    memset(sqTimeSlotMask, 0xFF, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    
    // This TS is already connected in the code above
    nTs = 0;
    sqTimeSlotMask[nTs] = 0x00;
    
    result = OtxDrvCreatePipe(2048-64, &hPipe);
    DemoCheckRc(hPipe, result);
    
    // -------------------------
    // Connect pipe: Dsp -> Burst
    // -------------------------
    result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(hPipe, result);
    
    result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hBurst, 0, sqTimeSlotMask, 32);
    DemoCheckRc(hPipe, result);
    
  }
  */
#endif


}


void DemoInitializeLogicalDevices(
/*******************************************************************************
  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.
*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  DemoLogf("In DemoInitializeLogicalDevices\n");


  DemoLogf("Configuring Lis for country set=");
  switch (g_demoEnv.nSetId) {
  case OTX_CST_POTS_CFG_SET_USA:
    DemoLogf("USA");
    break;
  case OTX_CST_POTS_CFG_SET_GERMANY: 
    DemoLogf("Germany");
    break;
  case OTX_CST_POTS_CFG_SET_UK:
    DemoLogf("United Kingdom");
    break;
  case OTX_CST_POTS_CFG_SET_FRANCE:
    DemoLogf("France");
    break;
  case OTX_CST_POTS_CFG_SET_AUSTRALIA:
    DemoLogf("Australia");
    break;
  case OTX_CST_POTS_CFG_SET_JAPAN:
    DemoLogf("Japan");
    break;
  default:
    DemoLogf("UNKNOWN");
    break;
  }
  DemoLogf("\n");  

  for (i = 0; i < DEMO_LI_MAX; i++) {

    // Initialize the POTS LI logical devices
    result = OtxCstPotsInit(
      g_hLogicDevs.hPotsLi[i],
      g_demoEnv.law,
      g_demoEnv.nSetId
      );
    DemoCheckRc(g_hLogicDevs.hPotsLi[i], result);


    // Initialize the Voice Activity Detector
    result = OtxCstVoiceActivityDetectorInit(
      g_hLogicDevs.hVaDetector[i], 
      g_demoEnv.law
      ); 
    DemoCheckRc(g_hLogicDevs.hVaDetector[i], result);

    // Initialize the CID Detector
    result = OtxCstCidDetectorInit(
      g_hLogicDevs.hCidDetector[i], 
      g_demoEnv.law, 
      OTX_CST_CID_DETECTOR_STD_CFG
      ); 
    DemoCheckRc(g_hLogicDevs.hCidDetector[i], result);

    // Initialize the CPT Detector
    result = OtxCstCptDetectorInit(
      g_hLogicDevs.hCptDetector[i], 
      g_demoEnv.law, 
      OTX_TRUE,
      OTX_CST_CPT_DETECTOR_Q35_CFG
      ); 
    DemoCheckRc(g_hLogicDevs.hCptDetector[i], result);

    // Initialize the DTMF Detector
    result = OtxCstDtmfDetectorInit(
      g_hLogicDevs.hDtmfDetector[i], 
      g_demoEnv.law, 
      OTX_TRUE); 
    DemoCheckRc(g_hLogicDevs.hDtmfDetector[i], result);

    // Initialize the DTMF Dialers
    result = OtxCstDtmfDialerInit(
      g_hLogicDevs.hDtmfDialer[i], 
      g_demoEnv.law 
      ); DemoCheckRc(g_hLogicDevs.hDtmfDialer[i], result);

    // Initialize the Voice Mixer
#if 0 // We can use either a simple Init function for the Voice Mixer (OtxCstVoiceMixerInit()),
      // where we do not specify the gains settings for each input or...
    result = OtxCstVoiceMixerInit(
      g_hLogicDevs.hVoiceMixer[i], 
      g_demoEnv.law, 
      2); 
#else // ... we can use a more advanced Init function for the Voice Mixer where we provide the 
      // gain setting for each input and the output.
    {
      OtxDataFormatE sqInDataFormats[2] = { g_demoEnv.law, g_demoEnv.law }; // Data Format of inputs 
      OtxDataFormatE eOutDataFormat      = g_demoEnv.law;     // Data Format of the output
      OTX_INT16 sqInDigGains[2]          = { 0 /* DTMF */, 0 /* Burst */ };        // Digital gain of the inputs (-1=-3dB, 0=0dB, 1=6dB, etc)
      OTX_INT16 nOutDigGain              = 0;                 // Digital gain of the output (-1=-3dB, 0=0dB, 1=6dB, etc).
      
      result = OtxCstVoiceMixerInitEx(
        g_hLogicDevs.hVoiceMixer[i], 
        sqInDataFormats,
        eOutDataFormat,
        sqInDigGains,
        nOutDigGain,
        2); 
    }
#endif
    DemoCheckRc(g_hLogicDevs.hVoiceMixer[i], result);

  }
}




void DemoEnableLogicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Logical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  DemoLogf("In DemoEnableLogicalDevices\n");

  for (i = 0; i < DEMO_LI_MAX; i++) {

    result = OtxDrvEnable(g_hLogicDevs.hPotsLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hPotsLi[i], result);

    result = OtxDrvEnable(g_hLogicDevs.hDtmfDetector[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hDtmfDetector[i], result);

    result = OtxDrvEnable(g_hLogicDevs.hCidDetector[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hCidDetector[i], result);

    result = OtxDrvEnable(g_hLogicDevs.hCptDetector[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hCptDetector[i], result);

    result = OtxDrvEnable(g_hLogicDevs.hVaDetector[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hVaDetector[i], result);

    result = OtxDrvEnable(g_hLogicDevs.hDtmfDialer[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hDtmfDialer[i], result);

    result = OtxDrvEnable(g_hLogicDevs.hVoiceMixer[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hVoiceMixer[i], result);

  }

}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX [%d.%d])\n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result));
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    DemoExit(1);
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
    
}


/*******************************************************************************
  @func
    Function that keeps reseting the Watchdog timer on the board
*******************************************************************************/
void  *DemoWatchDogFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_BOOL bEnabled;

  g_bWatchDogThreadActive = OTX_TRUE;
  g_bKickTheWatchDog = OTX_TRUE;

  DemoLogf("DemoWatchDogFunction Thread started\n");

  while (g_bWatchDogThreadActive) {

    result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  &bEnabled); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    if (!bEnabled) {
      DemoLogf("*********** Watchdog Timer timed out. The PC must be rebooted *********\n");
      OsaThreadSleep(3000); // in milliseconds
      continue;
    }

    if (g_bKickTheWatchDog) {
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_RESET,  OTX_TRUE /* Any value will work */); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
      OsaThreadSleep(100); // in milliseconds
    }
  }

  DemoLogf("DemoWatchDogFunction Thread terminating\n");
  return 0;

}

/*******************************************************************************
  @func
    Function that keeps monitoring the Loop current for Far-End-Hangup conditions
*******************************************************************************/
void  *DemoLineSensingFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_INT32  nCurrentLi = 0;

  g_bLineSensingThreadActive = OTX_TRUE;

  DemoLogf("DemoLineSensingFunction Thread started\n");
  while (g_bLineSensingThreadActive) {

    if (!g_bLineSensePending) {
      result = OtxCstPotsSenseLine(g_hLogicDevs.hPotsLi[nCurrentLi], 0x1006, 300); 
      if (OTX_SUCCESS(result)) {
        g_bLineSensePending = OTX_TRUE;
      }

    }

    OsaThreadSleep(100); // in milliseconds
  }

  DemoLogf("DemoLineSensingFunction Thread terminating\n");
  return 0;

}



/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;

}


/*******************************************************************************
  @func
    Only used by the USER_APPLICATION device for debugging of DSP program code.
*******************************************************************************/
void DemoPrintCSTEvent(OtxEventDataS *pEventData) 
{
  OTX_RESULT otxRc;
  char msgStr[256];
  
  if (pEventData->m_nCode == OTX_USER_APPLICATION_EC_READ_DATA) {
    OTX_UINT8  buf[8192];
    OTX_UINT32 nBytesReturned;
    OTX_UINT32 nIoControlCode;      
    
    otxRc = OtxDspReadData(pEventData->m_hDevice, OTX_TASK_SYNC, buf, 8192, &nBytesReturned, &nIoControlCode);
    if (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
      switch(nIoControlCode) {
      case 0x2ff:
        {
          char strBuf[1024];
          int k;
          sprintf(strBuf, "Len=%d :",nBytesReturned);
          for(k=0;k<(int)nBytesReturned;k+=2){
            sprintf(&strBuf[0] + strlen(strBuf), "%c",buf[k]);
          }
          sprintf(&strBuf[0] + strlen(strBuf), "\n");
          DemoLogf(strBuf);
        }
        break;
        
      case 0x0200 /*DEMO1_NOTIFY_DEBUG_DATA*/:
        {
          DemoLogf("DEMO1_NOTIFY_DEBUG_DATA (nRequestId=%d, nBytes=%d)\n", pEventData->m_nRequestId, nBytesReturned);
        }
        break;

      case 0x02FE:
        {
          static int counter = 0;

          if (counter++ % 5000 == 0) {
            DemoLogf("Got data\n");
          }

          if (g_fStreamFile) {
            fwrite((OTX_UINT8 *)buf, nBytesReturned, sizeof(OTX_UINT8), g_fStreamFile);
          }

        }
        break;
      default:
        DemoLogf("Unknown IoControlCode received in OtsDspReadData (nIoControlCode = %d, nBytesReturned=%d, nRequestId=%d)\n", nIoControlCode, nBytesReturned, pEventData->m_nRequestId);
        break;
      }
    } else {
      OTX_RESULT errorCode = otxRc;
      otxRc = OtxDrvResultCode2String(pEventData->m_hDevice, errorCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("User App[%d] Error: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("User App[%d] Invalid Error Code: %ld\n", pEventData->m_nSourceId, errorCode);
      }
      
    }
    
  } else {
    
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("User App[%d] Event: %s (nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nRequestId);
    } else {
      DemoLogf("User App[%d] Invalid Event: %ld (nRequestId=%d)\n", pEventData->m_nSourceId, pEventData->m_nCode, pEventData->m_nRequestId);
    }
  }
}




/*******************************************************************************
  @func
    Burst Sender Function
*******************************************************************************/
void DemoBurstSender(OTX_UINT32 nSourceId)
{
  OTX_RESULT  otxRc;
  OTX_UINT32  ui32;


  // validate parameters
  if (nSourceId != 0) {
    DemoLogf("DemoBurstSender: Invalid nSourceId(%d)!\n", nSourceId);
    return;
  }

  if (!g_sBurstDevs.mBurstSender.bActive) {
    // we're not active, so just return
    return;
  }

  if (g_sBurstDevs.mBurstSender.pTxFile) {
    // Prime the transmit buffer with 0xFF (only really needed for single TS transmission)
    memset(&g_sBurstDevs.mBurstSender.sqTxBuf[0], 0xFF, sizeof(g_sBurstDevs.mBurstSender.sqTxBuf) / sizeof(g_sBurstDevs.mBurstSender.sqTxBuf[0]));

    if (g_sBurstDevs.mBurstSender.nTsToSend == -1) {
      // All TS
      g_sBurstDevs.mBurstSender.cbTxBuf = fread(
        &g_sBurstDevs.mBurstSender.sqTxBuf,
        sizeof(g_sBurstDevs.mBurstSender.sqTxBuf[0]),
        sizeof(g_sBurstDevs.mBurstSender.sqTxBuf) / sizeof(g_sBurstDevs.mBurstSender.sqTxBuf[0]),
        g_sBurstDevs.mBurstSender.pTxFile
      );

    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE / 32];
      int cbSingleTsBuf;
      int byteCount = 0;
      int x;
      
      cbSingleTsBuf = fread(
        &singleTsBuf,
        sizeof(singleTsBuf[0]),
        sizeof(singleTsBuf) / sizeof(singleTsBuf[0]),
        g_sBurstDevs.mBurstSender.pTxFile
        );

      
      // Dispurse the single TS to be transmitted in the whole transmit buffer 
      byteCount = 0;
      for ( x = g_sBurstDevs.mBurstSender.nTsToSend; 
      (byteCount < cbSingleTsBuf) && (x < DEMO_BURST_BUF_SIZE); 
      x += 32
        ) {
        g_sBurstDevs.mBurstSender.sqTxBuf[x] = singleTsBuf[byteCount++];
      }
      
      assert(byteCount == cbSingleTsBuf);
      g_sBurstDevs.mBurstSender.cbTxBuf = cbSingleTsBuf * 32;

      if (!feof(g_sBurstDevs.mBurstSender.pTxFile)) {
        assert(g_sBurstDevs.mBurstSender.cbTxBuf == DEMO_BURST_BUF_SIZE);
      } else {
       DemoLogf("DemoBurstSender[%d]: Reached EOF\n", nSourceId);
      }

    }
        
    if (g_sBurstDevs.mBurstSender.cbTxBuf > 0) {

      //DemoLogf("DemoBurstSender[%d]: Transmitting %d bytes\n", nSourceId, g_sBurstDevs.mBurstSender.cbTxBuf);

      ui32 = g_sBurstDevs.mBurstSender.cbTxBuf;
      otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_sBurstDevs.mBurstSender.sqTxBuf, &ui32);
      DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

      // check the amount buffered w/ how much we requested to be buffered
      //  (since we set the trigger amount to the buffer size, there should
      //    have been room to buffer this, so these two should be the same).
      assert(ui32 == g_sBurstDevs.mBurstSender.cbTxBuf);

    } else {

      if (g_sBurstDevs.mBurstSender.bFileLoop) {
        if (g_sBurstDevs.mBurstSender.bVerbose) {
          DemoLogf("DemoBurstSender[%d]: Transmission complete. Starting file from the beginning.\n", nSourceId);
        }
        rewind(g_sBurstDevs.mBurstSender.pTxFile);
      } else {
        DemoLogf("DemoBurstSender[%d]: Transmission complete. Closing file\n", nSourceId);
        fclose(g_sBurstDevs.mBurstSender.pTxFile);
        g_sBurstDevs.mBurstSender.pTxFile = NULL;
        g_sBurstDevs.mBurstSender.bActive = OTX_FALSE;
      }
    }

  } else {
    // Send a text string (continously until aborted)

    // we MUST round amount to send to be bounded on the number of TS the highway hosts.
    //  (the buffer is representative of a burst DMA buffer for all TS on the highway hosted, TS0-TSn)

    g_sBurstDevs.mBurstSender.cbTxBuf = g_sBurstDevs.mBurstSender.cbTxBuf;
    assert(g_sBurstDevs.mBurstSender.cbTxBuf == DEMO_BURST_BUF_SIZE);

    ui32 = g_sBurstDevs.mBurstSender.cbTxBuf;
    otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_sBurstDevs.mBurstSender.sqTxBuf, &ui32);
    DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

    // check the amount buffered w/ how much we requested to be buffered
    //  (since we set the trigger amount to the buffer size, there should
    //    have been room to buffer this, so these two should be the same).
    assert(ui32 == g_sBurstDevs.mBurstSender.cbTxBuf);
  }

  return;
}

/*******************************************************************************
  @func
    Burst Receiver Function
*******************************************************************************/
void DemoBurstReceiver(OTX_UINT32 nSourceId)
{
  OTX_RESULT    otxRc;
  OTX_UINT8     buf[DEMO_BURST_BUF_SIZE];
  OTX_UINT32    ui32;
  OTX_UINT64    nFrameRefId;
  OTX_DATETIME  dtFrameRefTime;

  // validate parameters
  if (nSourceId < 0 || nSourceId > 2) {
    DemoLogf("DemoBurstReceiver: Invalid nSourceId(%d)...\n", nSourceId);
    return;
  }

  if (!g_sBurstDevs.mBurstReceiver.bActive) {
    // we're not active, so just return
    return;
  }

  // we want the frame reference id and time
  ui32 = sizeof(buf);
  otxRc = OtxBurstReadDataEx(g_hPhysDevs.hBurst, nSourceId, buf, &ui32, &nFrameRefId, &dtFrameRefTime);
  DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

  // check the amount buffered w/ how much we requested to be buffered
  //  (since we set the trigger amount to the buffer size, there should
  //    have been room to buffer this, so these two should be the same).
  assert(ui32 == sizeof(buf));

  // echo if we're in verbose mode
  if (g_sBurstDevs.mBurstReceiver.bVerbose) {
    int x;

    DemoPrintTimeStamp(dtFrameRefTime);

    // note: we only display the least significant 32 bits of the frame reference id.
    DemoLogf("BurstReceiver[%d] (len=%.5d) ID:%08x ", nSourceId, sizeof(buf), (OTX_UINT32) (nFrameRefId & 0xffffffff));

    for (x = 0; (x < 32) && (x < (OTX_INT32) sizeof(buf)); x++) {
      // display the highway's currentLi's data
      if ((x % 8) == 0) DemoLogf("\n");
      DemoLogf("%.2X ", buf[x]);
    }
    DemoLogf("\n");
    DemoLogf("\n");
  }

  // write data to a file if desired
  if (g_sBurstDevs.mBurstReceiver.pRxFile) {
    int cbWritten;

    if (g_sBurstDevs.mBurstReceiver.nTsToRecord == -1) {
      // All TS
      cbWritten = fwrite(buf, sizeof(buf[0]), sizeof(buf), g_sBurstDevs.mBurstReceiver.pRxFile);
      if (cbWritten != (int) sizeof(buf)) {
        DemoLogf("BurstReceiver[%d] Error: Could not write data to file.  (nFrameLen=%d, cbWritten=%d) \n", nSourceId, sizeof(buf), cbWritten);
      }
    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE/32];
      int byteCount = 0;
      int x;

      byteCount = 0;
      for (x = g_sBurstDevs.mBurstReceiver.nTsToRecord; x < (OTX_INT32) sizeof(buf); x += 32) {
        singleTsBuf[byteCount++] = buf[x];
      }
      cbWritten = fwrite(singleTsBuf, sizeof(singleTsBuf[0]), byteCount, g_sBurstDevs.mBurstReceiver.pRxFile);
      if (cbWritten != byteCount) {
        DemoLogf("BurstReceiver[%d] Error: Could not write data to file.  (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
      }
    }
  }
}


/*******************************************************************************

  @func

  Convert a CPT (Call Progress Tone) Detection parameter to a string.

  @comm

*******************************************************************************/
static char *DemoCptParamToStr(int nParam) 
{
  static char strBuf[128];

  switch (nParam) {

  case OTX_CST_CPT_DETECTOR_NO_TONE:        
    return "Silence, No tone";
  case OTX_CST_CPT_DETECTOR_DIAL_TONE:      
    return "Dial Tone";
  case OTX_CST_CPT_DETECTOR_RING_TONE:      
    return "Ring Tone";
  case OTX_CST_CPT_DETECTOR_BUSY_TONE:      
    return "Busy Tone";
  case OTX_CST_CPT_DETECTOR_FASTBUSY_TONE:  
    return "FastBusy Tone";
  case OTX_CST_CPT_DETECTOR_LONG_TONE:      
    return "Long lasting tone for 1 second";
  case OTX_CST_CPT_DETECTOR_END_OF_TONE:    
    return "End of current tone";
  default: // Error
    // DemoLogf("Unrecognized CPT tone parameter (%d)\n", nParam);
    sprintf(strBuf, "Unrecognized CPT Parameter (nParam=0x%X)", nParam);
    return strBuf;
  }
}



void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  //DemoPrintTimeStamp(pEventData->m_nTimeStamp);

  switch (pEventData->m_eDeviceType) {
    
  case OTX_LDEVICE_USER_APPLICATION:
    // Events from the DspDemo1.out logical devices
    DemoPrintCSTEvent(pEventData);
    break;

  case OTX_DEVICE_DSP:
    {
      // Non-logical device related event from the DSPs
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("DSP[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
    
  case OTX_LDEVICE_CST_DTMF_DIALER:
    {
      switch (pEventData->m_nCode) {
      case OTX_CST_DTMF_DIALER_EC_COMPLETE:
        DemoLogf("DtmfDialer[%d] Completed DTMF dialing of %d digits\n", pEventData->m_nSourceId, pEventData->m_nParam);
        break;
      default:
        DemoLogf("DtmfDialer[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        break;
      }
    }
    break;


  case OTX_LDEVICE_CST_CPT_DETECTOR:
    {
      // Event from Cpt Detector
      switch (pEventData->m_nCode) {
      case OTX_CST_CPT_DETECTOR_EC_MTONESTART:
        {
          DemoLogf("CptDetector[%d] MTONESTART, %s\n", pEventData->m_nSourceId, DemoCptParamToStr(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_CPT_DETECTOR_EC_MTONE:
        {
          DemoLogf("CptDetector[%d] MTONE, Tone confirmed, %s\n", pEventData->m_nSourceId, DemoCptParamToStr(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_CPT_DETECTOR_EC_MSIGNALEND:
        {
          DemoLogf("CptDetector[%d] MSIGNALEND, %s\n", pEventData->m_nSourceId, DemoCptParamToStr(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_CPT_DETECTOR_EC_MFAIL:
        {
          DemoLogf("CptDetector[%d] MFAIL, %s\n", pEventData->m_nSourceId, DemoCptParamToStr(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_CPT_DETECTOR_EC_MPATTERNEND:
        {
          DemoLogf("CptDetector[%d] MPATTERNEND, %s\n", pEventData->m_nSourceId, DemoCptParamToStr(pEventData->m_nParam));
        }
        break;
        
      default:
        DemoLogf("CptDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        break;
        
      }
    }
    break;

  case OTX_LDEVICE_CST_DTMF_DETECTOR:
    {
      // Event from Dtmf Detector
      switch (pEventData->m_nCode) {
      case OTX_CST_DTMF_DETECTOR_EC_MTONESTART:
        {
          DemoLogf("DtmfDetector[%d] MTONESTART, DTMF digit %c\n", pEventData->m_nSourceId, DemoDtmfParamToChar(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_DTMF_DETECTOR_EC_MTONE:
        {
          DemoLogf("DtmfDetector[%d] MTONE, DTMF tone confirmed\n", pEventData->m_nSourceId);
        }
        break;
        
      case OTX_CST_DTMF_DETECTOR_EC_MSIGNALEND:
        {
          DemoLogf("DtmfDetector[%d] MSIGNALEND, DTMF digit %c\n", pEventData->m_nSourceId, DemoDtmfParamToChar(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_DTMF_DETECTOR_EC_MFAIL:
        {
          DemoLogf("DtmfDetector[%d] MFAIL\n", pEventData->m_nSourceId);
        }
        break;
        
      case OTX_CST_DTMF_DETECTOR_EC_MPATTERNEND:
        {
          DemoLogf("DtmfDetector[%d] MPATTERNEND\n", pEventData->m_nSourceId);
        }
        break;
        
      default:
        DemoLogf("DtmfDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        break;
        
      }
    }
    break;

  case OTX_LDEVICE_CST_VOICE_ACTIVITY_DETECTOR:
    {
      switch (pEventData->m_nCode) {
      case OTX_CST_VOICE_ACTIVITY_DETECTOR_EC_NOISE:
        DemoLogf("VoiceActivityDetector[%d] Noise present\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_VOICE_ACTIVITY_DETECTOR_EC_NOISE_HANGOVER:
        DemoLogf("VoiceActivityDetector[%d] Noise Hangover. Possibly turning into voice.\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_VOICE_ACTIVITY_DETECTOR_EC_VOICE:
        DemoLogf("VoiceActivityDetector[%d] Voice present\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_VOICE_ACTIVITY_DETECTOR_EC_VOICE_HANGOVER:
        DemoLogf("VoiceActivityDetector[%d] Voice Hangover. Had voice, but possibly turning into noice.\n", pEventData->m_nSourceId);
        break;
      default:
        DemoLogf("VoiceActivityDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        break;
      }
    }
    break;


  case OTX_LDEVICE_CST_LI_POTS:
    {
      // Event from Pots Li
      switch (pEventData->m_nCode) {

      case OTX_CST_POTS_RING_SIGNAL_DETECTED:
        DemoLogf("PotsLi[%d] A ring signal has been detected\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_END_RING_SIGNAL_DETECTED:
        DemoLogf("PotsLi[%d] End of ring signal has been detected\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_AUTO_END_RING_SIGNAL_DETECTED:
        DemoLogf("PotsLi[%d] End of ring signal due to offhook has been detected\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_TIP_RING_REVERSAL_DETECTED:
        DemoLogf("PotsLi[%d] Tip/Ring reversal detected\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_CALLER_ID_DETECTED:
        {
          char szCallerId[40];
          OtxCstCidPresentationMessageS sCidMsg;
          OTX_UINT32 cbAttribValue;
          
          if (pEventData->m_nParam == OTX_CST_CALLER_ID_EC_NOERROR) {
            otxRc = OtxDrvGetAttributeValueSz(pEventData->m_hDevice, OTX_ATTR_CST_POTS_CALLER_ID, szCallerId, 40); 
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf("PotsLi[%d] Caller Id = \"%s\"\n", pEventData->m_nSourceId, szCallerId);
            } else if (OTX_RESULT_CODE(otxRc) == OTX_RES_BUSY) {
              DemoLogf("PotsLi[%d] Caller Id Overrun. Caller Id string lost. Reseting the condition\n", pEventData->m_nSourceId);
              OtxDrvSetAttributeValueSz(pEventData->m_hDevice, OTX_ATTR_CST_POTS_CALLER_ID, NULL); 
            } else {
              DemoLogf("PotsLi[%d] Unexpected return code from OtxDrvGetAttributeValueSz (otxRc=0x%X)\n", pEventData->m_nSourceId, otxRc);
            }
          } else {
            DemoLogf("PotsLi[%d] Caller Id detection failure. nCode=%d\n", pEventData->m_nSourceId, pEventData->m_nParam);
          }
          
          otxRc = OtxDrvGetAttributeValue(pEventData->m_hDevice,OTX_ATTR_CST_POTS_CALLER_ID_PRESENTATION_MESSAGE, &sCidMsg, sizeof(sCidMsg), &cbAttribValue); 
          if (OTX_SUCCESS(otxRc)) {
            DemoLogf("PotsLi[%d] Caller Id Presentation Message:\n", pEventData->m_nSourceId);
            DemoPrintPresentationMessage(&sCidMsg);
          } else if (OTX_RESULT_CODE(otxRc) == OTX_RES_BUSY) {
            DemoLogf("PotsLi[%d] Caller Id Overrun. Caller Msg lost. Reseting the condition\n", pEventData->m_nSourceId);
            OtxDrvSetAttributeValue(pEventData->m_hDevice, OTX_ATTR_CST_POTS_CALLER_ID_PRESENTATION_MESSAGE, NULL, 0); 
          } else {
            DemoLogf("PotsLi[%d] Unexpected return code from OtxDrvGetAttributeValue (otxRc=0x%X)\n", pEventData->m_nSourceId, otxRc);
          }
          
          
        }
        break;
      case OTX_CST_POTS_OFF_HOOK_COMPLETE:
        assert(pEventData->m_nSourceId < DEMO_LI_MAX);
        g_sPotsDevs[pEventData->m_nSourceId].mPotsLi.bOffHook = OTX_TRUE;
        DemoLogf("PotsLi[%d] Off-Hook command complete\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_ON_HOOK_COMPLETE:
        assert(pEventData->m_nSourceId < DEMO_LI_MAX);
        g_sPotsDevs[pEventData->m_nSourceId].mPotsLi.bOffHook = OTX_FALSE;
        DemoLogf("PotsLi[%d] On-Hook command complete\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_ON_HOOK_LINE_MON_COMPLETE:
        assert(pEventData->m_nSourceId < DEMO_LI_MAX);
        g_sPotsDevs[pEventData->m_nSourceId].mPotsLi.bOffHook = OTX_FALSE;
        DemoLogf("PotsLi[%d] On-Hook Line Monitor mode command complete\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_SENSE_LINE_COMPLETE:
        {
          double nVal = 2.75 * pEventData->m_nParam;
        
          assert(pEventData->m_nSourceId < DEMO_LI_MAX);
          if (g_bLineSenseVerbose) {
            //DemoPrintTimeStamp(pEventData->m_nTimeStamp);
          }
          if (g_sPotsDevs[pEventData->m_nSourceId].mPotsLi.bOffHook) {
            // We are in the OffHook state so line sensing will return loop current (3mA per bit)
            if (g_bLineSenseVerbose) {
              DemoLogf("PotsLi[%d] Loop current = %d mA (%d) above minimum\n", pEventData->m_nSourceId, 3 * pEventData->m_nParam, pEventData->m_nParam);
            }

            // Look for Far-End-Hangup. Two consecutive loop current values of zero is an indication of a Far-End-Hangup.
            if ((g_nLineSenseLastValue == 0) && (pEventData->m_nParam == 0)) {
              OTX_RESULT result;
              DemoLogf("PotsLi[%d] Detected Far-End-Hangup. Going OnHook.\n", pEventData->m_nSourceId);
            
              result = OtxCstPotsOnHook(pEventData->m_hDevice, 0x3001, 500); 
              if (!OTX_SUCCESS(result)) {
                DemoLogf("PotsLi[%d] On-Hook command failed\n", pEventData->m_nSourceId);
              }

            }
          } else {
            // We are in the OnHook state so line sensing will return line voltage (2.75V per bit)
            if (g_bLineSenseVerbose) {
              DemoLogf("PotsLi[%d] Line voltage = %2.1f V  (%d)\n", pEventData->m_nSourceId, (double) nVal, pEventData->m_nParam);
            }
          }
          DemoLogf("PotsLi[%d] Line voltage = %2.1f V  (%d)\n", pEventData->m_nSourceId, (double) nVal, pEventData->m_nParam);
          g_nLineSenseLastValue = pEventData->m_nParam;
        }
        break;
      case OTX_CST_POTS_PULSE_GEN_COMPLETE:
        DemoLogf("PotsLi[%d] Pulse Generator command complete\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_READ_REG_COMPLETE:
        DemoLogf("PotsLi[%d] ReadReg command complete Val=0x%.2X\n", pEventData->m_nSourceId, pEventData->m_nParam);
        break;
      case OTX_CST_POTS_WRITE_REG_COMPLETE:
        DemoLogf("PotsLi[%d] WriteReg command complete\n", pEventData->m_nSourceId);
        break;
      default:
        // Unexpected event from this device
        DemoLogf("PotsLi[%d] Unexpected Event (%d)\n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
    
  case OTX_LDEVICE_CST_CALLER_ID_DETECTOR:
    {
      // Event from Caller Id detector
      switch (pEventData->m_nCode) {

      case OTX_CST_CID_DETECTOR_DETECTED_CID:
        {
          char szCallerId[40];
          OtxCstCidPresentationMessageS sCidMsg;
          OTX_UINT32 cbAttribValue;

          if (pEventData->m_nParam == OTX_CST_CALLER_ID_EC_NOERROR) {
            otxRc = OtxDrvGetAttributeValueSz(pEventData->m_hDevice, OTX_ATTR_CST_CID_DETECTOR_CID_STRING, szCallerId, 40); 
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf("Cid[%d] Caller Id = \"%s\"\n", pEventData->m_nSourceId, szCallerId);
            } else if (OTX_RESULT_CODE(otxRc) == OTX_RES_BUSY) {
              DemoLogf("Cid[%d] Caller Id Overrun. Caller Id string lost. Reseting the condition\n", pEventData->m_nSourceId);
              OtxDrvSetAttributeValueSz(pEventData->m_hDevice, OTX_ATTR_CST_CID_DETECTOR_CID_STRING, NULL); 
            } else {
              DemoLogf("Cid[%d] Unexpected return code from OtxDrvGetAttributeValueSz (otxRc=0x%X)\n", pEventData->m_nSourceId, otxRc);
            }


            otxRc = OtxDrvGetAttributeValue(pEventData->m_hDevice, OTX_ATTR_CST_CID_DETECTOR_PRESENTATION_MESSAGE, &sCidMsg, sizeof(sCidMsg), &cbAttribValue); 
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf("Cid[%d] Caller Id Presentation Message:\n", pEventData->m_nSourceId);
              DemoPrintPresentationMessage(&sCidMsg);
            } else if (OTX_RESULT_CODE(otxRc) == OTX_RES_BUSY) {
              DemoLogf("Cid[%d] Caller Id Overrun. Caller Msg lost. Reseting the condition\n", pEventData->m_nSourceId);
              OtxDrvSetAttributeValue(pEventData->m_hDevice, OTX_ATTR_CST_CID_DETECTOR_PRESENTATION_MESSAGE, NULL, 0); 
            } else {
              DemoLogf("Cid[%d] Unexpected return code from OtxDrvGetAttributeValue (otxRc=0x%X)\n", pEventData->m_nSourceId, otxRc);
            }


          } else {
            DemoLogf("Cid[%d] Caller Id detection failure. nCode=%d\n", pEventData->m_nSourceId, pEventData->m_nParam);
          }

        }
        break;
      default:
        // Unexpected event from this device
        DemoLogf("PotsLi[%d] Unexpected Event (%d)\n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;

  case OTX_DEVICE_BURST:
    {
      switch (pEventData->m_nCode) {
      case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
        DemoBurstReceiver(pEventData->m_nRequestId);
        break;
      case OTX_BURST_SENDER_EC_BUF_DEPLETED:
        DemoBurstSender(pEventData->m_nRequestId);
        break;
      default:
        otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
        if (OTX_SUCCESS(otxRc)) {
          DemoLogf("Burst[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
        } else {
          DemoLogf("Burst[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        }
        break;
      }
      
    }
    break;

  default:
    {
      // Unhandled event
      DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    }
    break;
    
  }
  
}


/*******************************************************************************

  @func

  Print a Caller ID Presentation Message

  @comm

*******************************************************************************/
static void DemoPrintPresentationMessage(OtxCstCidPresentationMessageS *pCidMsg) 
{
  OTX_INT32 i;

  if (pCidMsg->validFields & OTX_CID_VALIDDATE) {
    DemoLogf("date: %c%c/%c%c\n", pCidMsg->date[0], pCidMsg->date[1], pCidMsg->date[2], pCidMsg->date[3]);
    DemoLogf("time: %c%c:%c%c\n", pCidMsg->date[4], pCidMsg->date[5], pCidMsg->date[6], pCidMsg->date[7]);
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDCALLINGLINEIDENTITY) {
    DemoLogf("callingLineIdentity: ");
    for (i = 0; pCidMsg->callingLineIdentity[i] && (i < 21); i++) {
      DemoLogf("%c", pCidMsg->callingLineIdentity[i]);
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDCALLEDLINEIDENTITY) {
    DemoLogf("calledLineIdentity: ");
    for (i = 0; pCidMsg->calledLineIdentity[i] && (i < 21); i++) {
      DemoLogf("%c", pCidMsg->calledLineIdentity[i]);
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDCALLINGPARTYNAME) {
    DemoLogf("callingPartyName: ");
    for (i = 0; pCidMsg->callingPartyName[i] && (i < 51); i++) {
      DemoLogf("%c", pCidMsg->callingPartyName[i]);
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDFIRSTLINEIDENTITY) {
    DemoLogf("firstLineIdentity: ");
    for (i = 0; pCidMsg->firstLineIdentity[i] && (i < 21); i++) {
      DemoLogf("%c", pCidMsg->firstLineIdentity[i]);
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDREDIRECTEDNUMBER) {
    DemoLogf("redirectedNumber: ");
    for (i = 0; pCidMsg->redirectedNumber[i] && (i < 21); i++) {
      DemoLogf("%c", pCidMsg->redirectedNumber[i]);
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDCOMPLEMENTARYCALLINGLINEIDENTITY) {
    DemoLogf("complementaryCallingLineIdentity: ");
    for (i = 0; pCidMsg->complementaryCallingLineIdentity[i] && (i < 21); i++) {
      DemoLogf("%c", pCidMsg->complementaryCallingLineIdentity[i]);
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDREASONFORABSENCE) {
    DemoLogf("reasonForAbsence: ");
    switch (pCidMsg->reasonForAbsence) {
    case 'O':
      DemoLogf("'O' - Unavailable (probably out of range)");
      break;
    case 'P':
      DemoLogf("'P' - Private (CLIR involved) call");
      break;
    default:
      DemoLogf("Unknown %c (0x%.4X)", pCidMsg->reasonForAbsence, pCidMsg->reasonForAbsence);
      break;
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDREASONFORABSENCEOFPARTY) {
    DemoLogf("reasonForAbsenceOfParty: ");
    switch (pCidMsg->reasonForAbsenceOfParty) {
    case 'O':
      DemoLogf("'O' - Unavailable (probably out of range)");
      break;
    case 'P':
      DemoLogf("'P' - Private (CLIR involved) call");
      break;
    default:
      DemoLogf("Unknown %c (0x%.4X)", pCidMsg->reasonForAbsenceOfParty, pCidMsg->reasonForAbsenceOfParty);
      break;
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDCALLTYPE) {
    DemoLogf("callType: ");
    switch (pCidMsg->callType) {
    case 0x01:
      DemoLogf("Voice Call");
      break;
    case 0x02:
      DemoLogf("CLI Ring Back in case of free call");
      break;
    case 0x03:
      DemoLogf("Calling Name Delivery");
      break;
    case 0x07:
      DemoLogf("Set Clock");
      break;
    case 0x81:
      DemoLogf("Message Waiting Call");
      break;
    default:
      DemoLogf("Unknown (0x%.4X)", pCidMsg->callType);
      break;
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDVISUALINDICATOR) {
    DemoLogf("visualIndicator: ");
    DemoLogf("%.4X ", pCidMsg->visualIndicator);
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDNETWORKMESSAGESYSTEMSTATUS) {
    DemoLogf("networkMessageSystemStatus: ");
    DemoLogf("%.4X ", pCidMsg->networkMessageSystemStatus);
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDEXTENSION) {
    DemoLogf("extension: ");
    for (i = 0; pCidMsg->extension[i] && (i < 10); i++) {
      DemoLogf("%c", pCidMsg->extension[i]);
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDTYPEOFFORWARDEDCALL) {
    DemoLogf("typeOfForwardedCall: ");
    switch (pCidMsg->typeOfForwardedCall) {
    case 0x01:
      DemoLogf("Call forwarded on busy");
      break;
    case 0x02:
      DemoLogf("Call forwarded on no reply");
      break;
    case 0x03:
      DemoLogf("Unconditional forwarded call");
      break;
    case 0x04:
      DemoLogf("Deflected call (after alerting)");
      break;
    case 0x05:
      DemoLogf("Deflected call (immediate)");
      break;
    case 0x06:
      DemoLogf("Call forwarded on inability to reach mobile subscriber");
      break;
    default:
      DemoLogf("Unknown (0x%.4X)", pCidMsg->typeOfForwardedCall);
      break;
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDTYPEOFCALLINGUSER) {
    DemoLogf("typeOfCallingUser: ");
    switch (pCidMsg->typeOfCallingUser) {
    case 0x00:
      DemoLogf("Origination unknown or unavailable");
      break;
    case 0x03:
      DemoLogf("Virtual Private Network");
      break;
    case 0x04:
      DemoLogf("Mobile phone");
      break;
    case 0x05:
      DemoLogf("Mobile phone + Virtual Private Network");
      break;
    case 0x0A:
      DemoLogf("Ordinary calling subscriber");
      break;
    case 0x0B:
      DemoLogf("Calling subscriber with priority");
      break;
    case 0x0C:
      DemoLogf("Data Call");
      break;
    case 0x0D:
      DemoLogf("Test Call");
      break;
    case 0x0F:
      DemoLogf("Payphone");
      break;
    default:
      DemoLogf("Unknown (0x%.4X)", pCidMsg->typeOfCallingUser);
      break;
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDLASTERROR) {
    DemoLogf("lastError: ");
    DemoLogf("%.4X ", pCidMsg->lastError);
    switch (pCidMsg->lastError) {
    case 0x00:
      DemoLogf("no error");
      break;
    case 0x10:
      DemoLogf("checksum error");
      break;
    case 0x11:
      DemoLogf("frame length error (unrecoverable)");
      break;
    case 0x12:
      DemoLogf("unrecognized message type (unrecoverable)");
      break;
    default:
      DemoLogf("Unknown (0x%.4X)", pCidMsg->lastError);
      break;
    }
    DemoLogf("\n");
  }
  
}

/*******************************************************************************

  @func

  Convert a DTMF parameter to a digit (char).

  @comm

*******************************************************************************/
static char DemoDtmfParamToChar(int nParam) {
  switch (nParam) {
  case 0:  // 0  (941Hz/1336Hz)
  case 1:  // 1  (697Hz/1209Hz)
  case 2:  // 2  (697Hz/1336Hz)
  case 3:  // 3  (697Hz/1477Hz)
  case 4:  // 4  (770Hz/1209Hz)
  case 5:  // 5  (770Hz/1336Hz)
  case 6:  // 6  (770Hz/1477Hz)
  case 7:  // 7  (852Hz/1209Hz)
  case 8:  // 8  (852Hz/1336Hz)
  case 9:  // 9  (852Hz/1477Hz)
    return '0' + nParam;
  case 10: // A  (697Hz/1633Hz)
  case 11: // B  (770Hz/1633Hz)
  case 12: // C  (852Hz/1633Hz)
  case 13: // D  (941Hz/1633Hz)
    return 'A' - 10 + nParam;
  case 14: // *  (941Hz/1209Hz)
    return '*';
    break;
  case 15: // #  (941Hz/1477Hz)
    return '#';
    break;
  default: // Error
    DemoLogf("Unrecognized DTMF tone parameter (%d)\n", nParam);
    return '-';
  }
}



/*******************************************************************************

  @func

    Returns the state of the Burst device (enabled or disabled).

*******************************************************************************/
OTX_BOOL DemoIsBurstDeviceEnabled()
{
  OTX_RESULT result;
  OtxDeviceStateE  eDevState;

  result = OtxDrvGetState(g_hPhysDevs.hBurst, &eDevState); 
  DemoCheckRc(g_hPhysDevs.hBurst, result);

  if (eDevState == OTX_DEVICE_WORKING) {
    return OTX_TRUE; // The burst device is enabled
  } else {
    return OTX_FALSE; // The burst device is not enabled
  }
}


OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************
  @func
    Processes User Commands
*******************************************************************************/
{
  OTX_RESULT        result;
  static OTX_INT32  nCurrentLi = 0;

  DemoLogf("'%c' pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd   Synopsis\n");
    DemoLogf("===   ===========================================================\n");
    DemoLogf(" ?    Display Help text\n");
    DemoLogf(" h    Ditto\n");
    DemoLogf(" q    Quit\n");

    //DemoLogf(" 0-5  Select active Line Interface\n");
    DemoLogf(" f    Set active Pots Li Off-Hook\n");
    DemoLogf(" n    Set active Pots Li On-Hook\n");
    DemoLogf(" m    Set active Pots Li in On-Hook Line-Monitor mode\n");
    DemoLogf(" c    Toggle verbose line sensing [Reading line voltage or loop current]\n");

    DemoLogf(" d    Call DTMF digits 102 on the active Pots Li\n");
    DemoLogf(" D    Dial user specified DTMF digits on the active Pots Li\n");
    DemoLogf(" l    Flash/hook on the active Pots Li\n");
    DemoLogf(" r    Read a Register from the DAA\n");
    DemoLogf(" w    Write to a Register in the DAA\n");
    DemoLogf(" b    Get Debug data from the DSP\n");
    DemoLogf(" s    Stream receive data to file (toggle). Only for DSP debugging\n");
    DemoLogf(" e    Enable/Disable Burst (toggle)\n");
    DemoLogf(" a    Abort Burst Senders and Receivers\n");
    DemoLogf(" R    Start Burst Receiver. Prompt for file to recv\n");
    DemoLogf(" S    Start Burst Sender. Prompt for file to send\n");
    DemoLogf(" L    Send loop. Start sending the file again once EOF is reached.\n");
    DemoLogf(" !    Terminate Watchdog Timer Thread (Crash test)\n");
    DemoLogf(" v    Toggle Burst Receiver Verbose mode\n");

    break;

  case 'v':
    {
      if (g_sBurstDevs.mBurstReceiver.bVerbose) {
        g_sBurstDevs.mBurstReceiver.bVerbose = OTX_FALSE;
      } else {
        g_sBurstDevs.mBurstReceiver.bVerbose = OTX_TRUE;
      }
      DemoLogf("Burst Receiver Verbose is now %s\n", g_sBurstDevs.mBurstReceiver.bVerbose ? "ON" : "OFF");
    }
    break;


  case '!':
    {
      DemoLogf("Terminating WatchDog Thread (Crash test)\n");
      g_bKickTheWatchDog = OTX_FALSE;
    }
    break;


  case 'e':
    if (DemoIsBurstDeviceEnabled()) {
      // by disabling the Burst device, we stop the process of transferring data
      // from the hardware to the driver via DMA.
      DemoLogf("Disable g_hPhysDevs.hBurst\n");
      result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
      DemoCheckRc(g_hPhysDevs.hBurst, result);
    } else {
      // by enabling the Burst device, we start the process of transferring data
      // from the hardware to the driver via DMA.  The data is stored in the Burst
      // device's buffer for later processing to connected logical devices.
      DemoLogf("Enable g_hPhysDevs.hBurst\n");
      result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
      DemoCheckRc(g_hPhysDevs.hBurst, result);
    }
    break;

  case 'a':
    {
      if (DemoIsBurstDeviceEnabled()) {
        if (g_sBurstDevs.mBurstReceiver.pRxFile) {
          DemoLogf("DemoBurstReceiver: Closing file\n");
          fclose(g_sBurstDevs.mBurstReceiver.pRxFile);
          g_sBurstDevs.mBurstReceiver.pRxFile = NULL;
        }
        g_sBurstDevs.mBurstReceiver.bActive = OTX_FALSE;
        
        if (g_sBurstDevs.mBurstReceiver.pRxFile) {
          DemoLogf("mBurstSender: Closing file\n");
          fclose(g_sBurstDevs.mBurstSender.pTxFile);
          g_sBurstDevs.mBurstSender.pTxFile = NULL;
        }
        g_sBurstDevs.mBurstSender.bActive   = OTX_FALSE;
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
    break;

  case 'R':
    {
      char sqzBuf[OTX_MAX_PATH];
      char sqzFile[OTX_MAX_PATH];
      static OTX_INT32 nTsToRecord = -1;

      if (DemoIsBurstDeviceEnabled()) {
        if (g_sBurstDevs.mBurstReceiver.bActive == OTX_FALSE || g_sBurstDevs.mBurstReceiver.pRxFile == NULL) {
          DemoLogf("BurstReceiver. Please enter file to receive: ");
          OsaConsoleGetStr(sqzFile);
          if (sqzFile[0] == '\0') {
            DemoLogf("Command aborted.\n");
            break;
          }
          
          if ((g_sBurstDevs.mBurstReceiver.pRxFile = fopen(sqzFile, "wb")) == NULL) {
            DemoLogf("Error: Could not open file %s. Command aborted.\n", sqzFile);
            break;
          }
          
          DemoLogf("BurstReceiver. Please enter single TS[0-31] to receive to file (-1=All) [%d]: ", nTsToRecord);
          OsaConsoleGetStr(sqzBuf);
          if (sqzBuf[0]) {
            nTsToRecord = strtol(sqzBuf, NULL, 0);
          }
          g_sBurstDevs.mBurstReceiver.nTsToRecord = nTsToRecord;
                         
          if (nTsToRecord == -1) {
            DemoLogf("Burst Receiver. Recording all 32 Burst Timeslots into file %s\n", sqzFile);
          } else {
            DemoLogf("Burst Receiver. Recording TS%d of Burst Highway into file %s\n", nTsToRecord, sqzFile);
          }
          
          g_sBurstDevs.mBurstReceiver.bActive = OTX_TRUE;
        } else {
          DemoLogf("Burst Receiver. Thread already running.\n");
        }
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
    break;

  case 'S':
    {
      char sqzBuf[OTX_MAX_PATH];
      static OTX_INT32 nTsToSend = 0;
      
      if (DemoIsBurstDeviceEnabled()) {
        if (g_sBurstDevs.mBurstSender.bActive == OTX_FALSE) {
          DemoLogf("BurstSender[%d]. Please enter file to send: ");
          
          OsaConsoleGetStr(sqzBuf);
          if (sqzBuf[0] == '\0') {
            break;
          }
          
          if ((g_sBurstDevs.mBurstSender.pTxFile = fopen(sqzBuf, "rb")) == NULL) {
            DemoLogf("Error: Could not open file %s \n", sqzBuf);
            break;
          }
          
          DemoLogf("BurstSender[%d]. Sending data from file on TS%d of the Burst highway\n", nTsToSend);
          
          g_sBurstDevs.mBurstSender.bActive = OTX_TRUE;
        } else {
          DemoLogf("Burst Sender[%d] already running.\n");
        }
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
    break;

  case 'L':
    {
      g_sBurstDevs.mBurstSender.bFileLoop = !g_sBurstDevs.mBurstSender.bFileLoop;
      DemoLogf("Setting loop file sends %s for Burst Sender[%d] Thread\n", (g_sBurstDevs.mBurstSender.bFileLoop)?"ON":"OFF", 0);
    }
    break;

  case 'b':
    {
      DemoLogf("Debug Dump for Dsp#%d\n", nCurrentLi);
      OtxDrvDbgDump(g_hPhysDevs.hDsp[nCurrentLi], 0);
    }
    break;


  case 'f':
    // Set Active Pots Li Off-Hook ...
    DemoLogf("Pots Li#%d Off-hook\n", nCurrentLi);
    result = OtxCstPotsOffHook(g_hLogicDevs.hPotsLi[nCurrentLi], 0x1000, 2000); 
    DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
    break;


  case 'n':
    // Set Active Pots Li On-Hook ...
    DemoLogf("Pots Li#%d On-hook\n", nCurrentLi);
    result = OtxCstPotsOnHook(g_hLogicDevs.hPotsLi[nCurrentLi], 0x1001, 2000); 
    DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
    break;

  case 'm':
    // Set Active Pots Li in On-Hook Line Monitor mode ...
    DemoLogf("Pots Li#%d On-hook Line-Monitor mode\n", nCurrentLi);
    result = OtxCstPotsOnHookLineMon(g_hLogicDevs.hPotsLi[nCurrentLi], 0x1001, 2000); 
    DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
    break;

  case 'c':
    // Line Sense
    g_bLineSenseVerbose = !g_bLineSenseVerbose;
    g_bLineSensePending = OTX_FALSE;
    DemoLogf("Pots Li#%d Line Sense is now %s\n", nCurrentLi, g_bLineSenseVerbose ? "Verbose" : "Silent");
    break;

  case 'l':
    // Flash Hook
    DemoLogf("Pots Li#%d Hook Flash\n", nCurrentLi);
    result = OtxCstPotsHookFlash(g_hLogicDevs.hPotsLi[nCurrentLi], 200, 0x1002, 2000); DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
    break;

  case 'd':
    {
      // Dial a hard-coded number ...
      OTX_HANDLE hHandle = g_hLogicDevs.hDtmfDialer[nCurrentLi];

      DemoLogf("Sending digits on Pots Li#%d: '%s'", nCurrentLi, g_sqNumber0);

      // Start Dialing DTMF number, Non-Blocking call, will receive Event when completed
      result = OtxCstDtmfDial(
        hHandle, 
        (unsigned char *)g_sqNumber0, 
        strlen(g_sqNumber0), 
        DEMO_TASK_REF, 
        OTX_TIME_INFINITY); 
      DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;

  case 'D':
    {
      // Dial a user specified number ...
      char strBuf[32];
      OTX_HANDLE hHandle = g_hLogicDevs.hDtmfDialer[nCurrentLi];

      DemoLogf("Phone number to Dial from Pots Li#%d: ", nCurrentLi);
      OsaConsoleGetStr(strBuf);
      DemoLogf("Pots Li#%d Dialling number '%s'", nCurrentLi, strBuf);

      // Start Dialing DTMF number, Non-Blocking call, will receive Event when completed
      result = OtxCstDtmfDial(
        hHandle, 
        (unsigned char *)strBuf, 
        strlen(strBuf), 
        DEMO_TASK_REF, 
        OTX_TIME_INFINITY); 
      DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;

  case 'p':
    {
      // Pulse dial
      OTX_INT32 c;
      DemoLogf("Pots Li#%d Pulse Dialing. Press digit to dial... [0-9] ", nCurrentLi);
      c = OsaConsoleGetCh();
      DemoLogf("\n");
      result = OtxCstPotsPulseDial(g_hLogicDevs.hPotsLi[nCurrentLi], c, 0x1003, 2000); DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
      DemoLogf("Done.\n");
    }
    break;

#if 0 // Only one DSP/FXO is supported by this program at this time
  /*
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
    // Select current Li
    {
      OTX_INT32 nLi = commandChar - '0';
      
      nCurrentLi = nLi;
      DemoLogf("Li#%d Selected ....\n", nCurrentLi);
    }
    break;
  */
#endif

#if 1 // Only for DSP program debugging
  case 's':
    {
      char sqzBuf[OTX_MAX_PATH];
      
      if (g_fStreamFile) {
        fclose(g_fStreamFile);
        g_fStreamFile = NULL;
        DemoLogf("Closed streaming file\n");
      } else {
        
        DemoLogf("Please enter file to receive: ");
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0] == '\0') {
          break;
        }
        
        if ((g_fStreamFile = fopen(sqzBuf, "wb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBuf);
          break;
        }
      }
    }
    break;
#endif

#if 1 // Only for DSP program debugging
  case 'r':
    {
      OTX_UINT32 nCmd = 1; // 1 = Read
      static OTX_UINT8 nRegister = 2; 
      static OTX_UINT8 nDummy2 = 0; 
      static OTX_UINT8 nDummy3 = 0; 
      char strBuf[64];
      
      DemoLogf("Read DAA register. Register [%d]\n ", nRegister);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nRegister = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }

      nCmd = (nDummy3 << 3*8) | (nDummy2 << 2*8) | (nRegister << 1*8) | nCmd;
      result = OtxDrvDbgDump(g_hLogicDevs.hPotsLi[nCurrentLi], nCmd);
      
    }
    break;
#endif

#if 1 // Only for DSP program debugging
  case 'w':
    {
      OTX_UINT32 nCmd = 2; // 2 = Write
      static OTX_UINT8 nRegister = 2; 
      static OTX_UINT8 nVal = 0; 
      static OTX_UINT8 nDummy3 = 0; 
      char strBuf[64];
      
      DemoLogf("Write DAA register. Register [%d]\n ", nRegister);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nRegister = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }

      DemoLogf("Please enter value [0x%.2X] ", nVal);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nVal = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }

      nCmd = (nDummy3 << 3*8) | (nVal << 2*8) | (nRegister << 1*8) | nCmd;
      result = OtxDrvDbgDump(g_hLogicDevs.hPotsLi[nCurrentLi], nCmd);
      
    }
    break;
#endif
  


  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;
      
      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);
      
      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
      
    }
    break;
    
  case 'q':
  case 'Q':
    // Quit ...
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}



/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->bExitOnError      = OTX_FALSE;
  envP->logFileName       = DEMO_DEFAULT_LOG_FILE;
  envP->outFileName       = NULL;
  envP->law               = OTX_DATA_FORMAT_COMPAND8_U_LAW;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->fxoNo             = DEMO_DEFAULT_FXO_NO;
  envP->nSetId            = OTX_CST_POTS_CFG_SET_USA;
  envP->nBoardType        = OTX_DEVICE_ARNI_6X6_PCI;
  envP->eDetectorType     = OTX_LDEVICE_CST_DTMF_DETECTOR;

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          envP->logFileName = argv[i]+2;

        } else if (*(argv[i]+1) == 'O') {   // DSP out file
          envP->outFileName = argv[i]+2;

        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);

        } else if (*(argv[i]+1) == 'F') {   // FXO number
          envP->fxoNo = strtoul(argv[i]+2, NULL, 0);

        } else if (OsaToUpper(*(argv[i]+1)) == 'T') {   // Board Type
          if (OsaToUpper(*(argv[i] + 2)) == 'A') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X6_PCI_PLUS;
          } else if (OsaToUpper(*(argv[i] + 2)) == 'J') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X6_PCI;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'C') {   // Companding Law
          if (*(argv[i]+2) == 'A') {   // A-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          } else if (*(argv[i]+2) == 'U') {   // U-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_U_LAW;
          } else {
            DemoLogf("Error: Invalid companding law command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'S') {   // Configuration Set
          if (*(argv[i]+2) == 'U') {          // USA
            envP->nSetId = OTX_CST_POTS_CFG_SET_USA;
          } else if (*(argv[i]+2) == 'G') {   // Germany
            envP->nSetId = OTX_CST_POTS_CFG_SET_GERMANY;
          } else if (*(argv[i]+2) == 'E') {   // UK
            envP->nSetId = OTX_CST_POTS_CFG_SET_UK;
          } else if (*(argv[i]+2) == 'F') {   // France
            envP->nSetId = OTX_CST_POTS_CFG_SET_FRANCE;
          } else if (*(argv[i]+2) == 'A') {   // Australia
            envP->nSetId = OTX_CST_POTS_CFG_SET_AUSTRALIA;
          } else if (*(argv[i]+2) == 'J') {   // Japan
            envP->nSetId = OTX_CST_POTS_CFG_SET_JAPAN;
          } else {
            DemoLogf("Error: Unsupported country specific configuration set %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'D') {   // Detector Type
          if (*(argv[i]+2) == 'C') {          // CPT
            envP->eDetectorType = OTX_LDEVICE_CST_CPT_DETECTOR;
          } else if (*(argv[i]+2) == 'D') {          // DTMF
            envP->eDetectorType = OTX_LDEVICE_CST_DTMF_DETECTOR;
          } else if (*(argv[i]+2) == 'V') {          // Voice Activity
            envP->eDetectorType = OTX_LDEVICE_CST_VOICE_ACTIVITY_DETECTOR;
          } else if (*(argv[i]+2) == 'I') {          // Voice Activity
            envP->eDetectorType = OTX_LDEVICE_CST_CALLER_ID_DETECTOR;
          } else {
            DemoLogf("Error: Unsupported detector type %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }

}


static void DemoHelpPanel()
/*******************************************************************************
  @func
    Displays a help for the command line.
*******************************************************************************/
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName>    : Log file name\n");
  DemoLogf(" /B<boardNo>     : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /F<fxoNo>       : FXO number to configure (Default is %d)\n", DEMO_DEFAULT_FXO_NO);
  DemoLogf(" /T[A|J]         : Board Type used for the demo, where\n");
  DemoLogf("                    A  = Use Arni-6-PCI-Plus Board\n");
  DemoLogf("                    J  = Use Arni-6-PCI Board\n");
  DemoLogf(" /D[D|C|V|I]     : Detector type, where\n");
  DemoLogf("                    D  = DTMF Detector (default)\n");
  DemoLogf("                    C  = Call Progress Tone Detector\n");
  DemoLogf("                    V  = Voice Activity Detector\n");
  DemoLogf("                    I  = Caller Id Detector\n");
  DemoLogf(" /C[A|U]         : Voice Compression Codec, where\n");
  DemoLogf("                    A  = A-law\n");
  DemoLogf("                    U  = u-law (default)\n");
  DemoLogf(" /S[U|G|E|F|A|J] : Country specific setting, where\n");
  DemoLogf("                    U  = Li Configuration for USA (default)\n");
  DemoLogf("                    G  = Li Configuration for Germany\n");
  DemoLogf("                    E  = Li Configuration for United Kingdom\n");
  DemoLogf("                    F  = Li Configuration for France\n");
  DemoLogf("                    A  = Li Configuration for Australia\n");
  DemoLogf("                    J  = Li Configuration for Japan\n");
  DemoLogf(" /O<fileName>    : DSP Out file name (for debugging only)\n");
}


static void DemoLogf(char *formatStr, ...)
/*******************************************************************************
  @func
    Displays a message on the screen and prints it to the logfile.
*******************************************************************************/
{
  va_list argList;

  OsaMutexLock(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  OsaMutexUnlock(&g_demoLogfCritSec);
}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  //OTX_INT32  i;
  OTX_INT32  exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char       commandChar;

	OsaConsoleInit();

  OsaMutexInit(&g_demoLogfCritSec);

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    exit(-1);
  }

  DemoLogf("%s, %s, Odin TeleSystems Inc.\n", argv[0], __DATE__);
  DemoLogf("Arni-6x6-PCI, Basic FXO Demo Program.\n");

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);
  DemoLogf("FXO   number %d\n", g_demoEnv.fxoNo);
  if (g_demoEnv.outFileName) {
    DemoLogf("Out file to be loaded is %s\n", g_demoEnv.outFileName);
  }

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    DemoExit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    DemoExit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  g_bLibConnected = OTX_TRUE;

  // Create Event Queues for Driver --> Application Communications
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Arni-16-PCI board
  DemoOpenPhysicalDevices();

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);

    free(pszCpldBuildNumbers);

  }

  // Enable Physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnablePhysicalDevices();

  OsaThreadStart(DemoIdleFunction);

  // Create and get handles to the Logical devices needed by this application
  DemoCreateLogicalDevices();

  // Create and get handles to the Pipe devices needed by this application
  // Connect Pipes to Physical and Logical Devices
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoInitializeLogicalDevices();

  // Enable Logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnableLogicalDevices();

#if 0 // Disable watchdog while debugging
  // Enable the Watchdog Timer itself
  result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  OTX_TRUE); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Start the Watchdog timer Reset Thread, which will keep the watchdog alive
  OsaThreadStart(DemoWatchDogFunction);
#endif  

  // Start the Line Sensing Thread, which requests period line sensing from the PotsLi to faciliate monitoring of Far-End-Hangup conditions
  OsaThreadStart(DemoLineSensingFunction);

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....

  DemoLogf("\nPress 'h' for help, 'q' to quit ...\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
		exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Delete allocated resources
  DemoExit(0);

  return 0;
}

/*******************************************************************************
*******************************************************************************/
static int DemoExit(int nExitCode)
{

  DemoLogf("\nExiting...\n");

  // Important to disable the Watchdog Timer since we will not be reseting it after we exit
  // the application. 
  OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  OTX_FALSE);

  g_bWatchDogThreadActive = OTX_FALSE;
  g_bLineSensingThreadActive = OTX_FALSE;

  if (DemoIsBurstDeviceEnabled()) {
    if (g_sBurstDevs.mBurstReceiver.pRxFile) {
      DemoLogf("DemoBurstReceiver: Closing file\n");
      fclose(g_sBurstDevs.mBurstReceiver.pRxFile);
      g_sBurstDevs.mBurstReceiver.pRxFile = NULL;
    }
    g_sBurstDevs.mBurstReceiver.bActive = OTX_FALSE;
    
    if (g_sBurstDevs.mBurstReceiver.pRxFile) {
      DemoLogf("mBurstSender: Closing file\n");
      fclose(g_sBurstDevs.mBurstSender.pTxFile);
      g_sBurstDevs.mBurstSender.pTxFile = NULL;
    }
    g_sBurstDevs.mBurstSender.bActive   = OTX_FALSE;
  }

  OsaThreadSleep(100);   // give threads a chance to exit

  // Clean up
  OsaConsoleExit();
  OtxDrvReleaseLib();

  exit(nExitCode);
}
